@@ -514,30 +514,26 @@ class Request(object):
 
     def __init__(self, environ=None, environ_getter=None, charset=NoDefault, unicode_errors=NoDefault,
                  decode_param_names=NoDefault, **kw):
-        if environ is None and environ_getter is None:
-            raise TypeError(
-                "You must provide one of environ or environ_getter")
-        if environ is not None and environ_getter is not None:
-            raise TypeError(
-                "You can only provide one of the environ and environ_getter arguments")
+        if environ_getter is not None:
+            raise ValueError('The environ_getter argument is no longer '
+                             'supported')
         if environ is None:
-            self._environ_getter = environ_getter
-        else:
-            if not isinstance(environ, dict):
-                raise TypeError(
-                    "Bad type for environ: %s" % type(environ))
-            self._environ = environ
+            raise TypeError("You must provide an environ arg")
+        d = self.__dict__
+        d['environ'] = environ
         if charset is not NoDefault:
-            self.__dict__['charset'] = charset
+            d['charset'] = charset
         if unicode_errors is not NoDefault:
-            self.__dict__['unicode_errors'] = unicode_errors
+            d['unicode_errors'] = unicode_errors
         if decode_param_names is not NoDefault:
-            self.__dict__['decode_param_names'] = decode_param_names
-        for name, value in kw.items():
-            if not hasattr(self.__class__, name):
-                raise TypeError(
-                    "Unexpected keyword: %s=%r" % (name, value))
-            setattr(self, name, value)
+            d['decode_param_names'] = decode_param_names
+        if kw:
+            my_class = self.__class__
+            for name, value in kw.iteritems():
+                if not hasattr(my_class, name):
+                    raise TypeError(
+                        "Unexpected keyword: %s=%r" % (name, value))
+                setattr(self, name, value)
 
     def __setattr__(self, attr, value, DEFAULT=[]):
         ## FIXME: I don't know why I need this guard (though experimentation says I do)
@@ -564,16 +560,6 @@ class Request(object):
         except KeyError:
             raise AttributeError(attr)
 
-    def environ(self):
-        """
-        The WSGI environment dictionary for this request
-        """
-        return self._environ_getter()
-    environ = property(environ, doc=environ.__doc__)
-
-    def _environ_getter(self):
-        return self._environ
-
     def _body_file__get(self):
         """
         Access the body of the request (wsgi.input) as a file-like
@@ -917,10 +903,9 @@ class Request(object):
             # Not an HTML form submission
             return NoVars('Not an HTML form submission (Content-Type: %s)'
                           % content_type)
-        if 'CONTENT_LENGTH' not in env:
-            # FieldStorage assumes a default CONTENT_LENGTH of -1, but a
-            # default of 0 is better:
-            env['CONTENT_TYPE'] = '0'
+        # FieldStorage assumes a default CONTENT_LENGTH of -1, but a
+        # default of 0 is better:
+        env.setdefault('CONTENT_LENGTH', 0)
         fs_environ = env.copy()
         fs_environ['QUERY_STRING'] = ''
         fs = cgi.FieldStorage(fp=self.body_file,
@@ -1440,7 +1425,7 @@ class Response(object):
     unicode_errors = 'strict'
     default_conditional_response = False
 
-    def __init__(self, body=None, status='200 OK', headerlist=None, app_iter=None,
+    def __init__(self, body=None, status=None, headerlist=None, app_iter=None,
                  request=None, content_type=None, conditional_response=NoDefault,
                  **kw):
         if app_iter is None:
@@ -1449,7 +1434,10 @@ class Response(object):
         elif body is not None:
             raise TypeError(
                 "You may only give one of the body and app_iter arguments")
-        self.status = status
+        if status is None:
+            self._status = '200 OK'
+        else:
+            self.status = status
         if headerlist is None:
             self._headerlist = []
         else:
@@ -1464,34 +1452,45 @@ class Response(object):
                 self._request = None
         else:
             self._environ = self._request = None
-        if content_type is not None:
-            self.content_type = content_type
-        elif self.default_content_type is not None and headerlist is None:
-            self.content_type = self.default_content_type
+        if content_type is None:
+            content_type = self.default_content_type
+        if 'charset' in kw:
+            charset = kw.pop('charset')
+        elif self.default_charset and headerlist is None:
+            if content_type and (content_type == 'text/html'
+                                 or content_type.startswith('text/')
+                                 or content_type.startswith('application/xml')
+                                 or (content_type.startswith('application/')
+                                     and content_type.endswith('+xml'))):
+                charset = self.default_charset
+        else:
+            charset = None
+        if content_type and charset:
+            content_type += '; charset=' + charset
+        elif self._headerlist and charset:
+            self.charset = charset
+        if not self._headerlist and content_type:
+            self._headerlist.append(('Content-Type', content_type))
         if conditional_response is NoDefault:
             self.conditional_response = self.default_conditional_response
         else:
             self.conditional_response = conditional_response
-        if 'charset' in kw:
-            # We set this early, so something like unicode_body works later
-            value = kw.pop('charset')
-            if value:
-                self.charset = value
-        elif self.default_charset and not self.charset and headerlist is None:
-            ct = self.content_type
-            if ct and (ct.startswith('text/') or ct.startswith('application/xml')
-                       or (ct.startswith('application/') and ct.endswith('+xml'))):
-                self.charset = self.default_charset
         if app_iter is not None:
             self._app_iter = app_iter
             self._body = None
         else:
             if isinstance(body, unicode):
-                self.unicode_body = body
+                if charset is None:
+                    raise TypeError(
+                        "You cannot set the body to a unicode value without a charset")
+                body = body.encode(charset)
+            self._body = body
+            if headerlist is None:
+                self._headerlist.append(('Content-Length', str(len(body))))
             else:
-                self.body = body
+                self.headers['Content-Length'] = str(len(body))
             self._app_iter = None
-        for name, value in kw.items():
+        for name, value in kw.iteritems():
             if not hasattr(self.__class__, name):
                 # Not a basic attribute
                 raise TypeError(
@@ -1511,6 +1510,19 @@ class Response(object):
                 + '\n\n'
                 + self.body)
 
+    def copy(self):
+        """Makes a copy of the response"""
+        if self._app_iter is not None:
+            app_iter = self._app_iter
+        else:
+            app_iter = [self._body]
+        return self.__class__(
+            content_type=False,
+            status=self._status,
+            headerlist=self._headerlist,
+            app_iter=app_iter,
+            conditional_response=self.conditional_response)
+
     def _status__get(self):
         """
         The status string
